#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <arpa/inet.h>  // htons()
#include <netinet/in.h> // struct sockaddr_in
#include <sys/socket.h>

#include "common.h"

void* connection_handler(void* arg) {
    /* Trick: pass an integer by converting it into a void. It will
     * come in handy for a multi-threaded version of the server. */
    int socket_desc = (int)arg;

    int ret, recv_bytes;

    char buf[1024];
    size_t buf_len = sizeof(buf);
    size_t msg_len;

    char* quit_command = SERVER_COMMAND;
    size_t quit_command_len = strlen(quit_command);

    // send welcome message
    sprintf(buf, "Hi! I'm an echo server.\nI will send you back whatever"
            " you send me. I will stop if you send me %s :-)\n", quit_command);
    msg_len = strlen(buf);
    /** INSERT CODE TO SEND WELCOME MESSAGE TO THE CLIENT
     *
     * Suggestions:
     * - send() with flags = 0 is equivalent to write() on a descriptor
     * - the message you have to send has already been written inside buf
     * - don't deal with messages partially sent
     */

    // echo loop
    while (1) {
        /** INSERT CODE TO RECEIVE A MESSAGE FROM THE CLIENT
         *
         * Suggestions:
         * - recv() with flags = 0 is equivalent to read() on a descriptor
         * - store the number of received bytes in recv_bytes
         * - on a socket we get 0 as return value only when the peer closes
         *   the connection: if there are no bytes to read and we invoke
         *   recv() we will get stuck since the call is blocking!
         */

        // check whether I have just been told to quit...
        /** INSERT CODE HERE TO ANALYZE THE RECEIVED MESSAGE
         *
         * Suggestions:
         * - compare the number of bytes received with the length of the
         *   quit command that tells the server to close the connection
         * - perform a byte-to-byte comparsion when required using
         *   memcmp(const void *ptr1, const void *ptr2, size_t num)
         * - exit from the cycle when there is nothing to send back
         */

        // ... or if I have to send the message back
        /** INSERT CODE TO ECHO THE RECEIVED MESSAGE BACK TO THE CLIENT
         *
         * Suggestions:
         * - send() with flags = 0 is equivalent to write() on a descriptor
         * - don't deal with messages partially sent
         */
    }

    // close socket
    ret = close(socket_desc);
    ERROR_HELPER(ret, "Cannot close socket for incoming connection");

    return NULL;
}

int main(int argc, char* argv[]) {
    int ret;

    int socket_desc, client_desc;

    // some fields are required to be filled with 0
    struct sockaddr_in server_addr = {0}, client_addr = {0};

    int sockaddr_len = sizeof(struct sockaddr_in); // we will reuse it for accept()

    // initialize socket for listening
    socket_desc = socket(AF_INET , SOCK_STREAM , 0);
    ERROR_HELPER(socket_desc, "Could not create socket");

    server_addr.sin_addr.s_addr = INADDR_ANY; // we want to accept connections from any interface
    server_addr.sin_family      = AF_INET;
    server_addr.sin_port        = htons(SERVER_PORT); // don't forget about network byte order!

    /* We enable SO_REUSEADDR to quickly restart our server after a crash:
     * for more details, read about the TIME_WAIT state in the TCP protocol */
    int reuseaddr_opt = 1;
    ret = setsockopt(socket_desc, SOL_SOCKET, SO_REUSEADDR, &reuseaddr_opt, sizeof(reuseaddr_opt));
    ERROR_HELPER(ret, "Cannot set SO_REUSEADDR option");

    // bind address to socket
    ret = bind(socket_desc, (struct sockaddr*) &server_addr, sockaddr_len);
    ERROR_HELPER(ret, "Cannot bind address to socket");

    // start listening
    ret = listen(socket_desc, MAX_CONN_QUEUE);
    ERROR_HELPER(ret, "Cannot listen on socket");

    // loop to handle incoming connections serially
    while (1) {
        client_desc = accept(socket_desc, (struct sockaddr*) &client_addr, (socklen_t*) &sockaddr_len);
        ERROR_HELPER(client_desc, "Cannot open socket for incoming connection");

        if (DEBUG) fprintf(stderr, "Incoming connection accepted...\n");

        /* Trick: we pass an int by converting it into a void*; the
         * will come in handy for a multi-threaded version of a server
         * as we rely on the pass-by-value semantics. */
        connection_handler((void*)client_desc);

        if (DEBUG) fprintf(stderr, "Done!\n");
    }

    exit(EXIT_SUCCESS); // this will never be executed
}
